"""
Utilities for writing and reading files compatible with Fortran

"""

import re

def f2s(f):
    """
    Format a string containing a float
    """
    s = ""
    if f >= 0.0:
        s += " "
    return s + "%1.9E" % f

class ChunkOutput:
    """
    This outputs values in lines, inserting
    newlines when needed.
    """
    def __init__(self, filehandle, chunksize=5, extraspaces=0):
        """
        filehandle  output to write to
        chunksize   number of values on a line
        extraspaces  number of extra spaces between outputs
        """
        self.fh = filehandle
        self.counter = 0
        self.chunk = chunksize
        self.extraspaces = extraspaces

    def write(self, value):
        """"
        Write a value to the output, adding a newline if needed
        """
        self.fh.write(" "*self.extraspaces)
        self.fh.write(f2s(value))
        self.counter += 1
        if self.counter == self.chunk:
            self.fh.write("\n")
            self.counter = 0

    def newline(self):
        """
        Ensure that the file is at the start of a new line
        """
        if self.counter != 0:
            self.fh.write("\n")
            self.counter = 0

    def endblock(self):
        """
        Make sure next block of data is on new line
        """
        self.fh.write("\n")
        self.counter=0
        
def write_1d(val, out):
    """
    Writes a 1D variable val to the file handle out
    """
    for i in range(len(val)):
        out.write(val[i])
    out.newline()
    

def write_2d(val, out):
    """
    Writes a 2D array. Note that this transposes
    the array, looping over the first index fastest
    """
    nx,ny = val.shape
    for y in range(ny):
        for x in range(nx):
            out.write(val[x,y])
    out.newline()

def next_value(fh):
    """
    A generator which yields values from a file handle
    
    Checks if the value is a float or int, returning
    the correct type depending on if '.' is in the string
    
    """
    pattern = re.compile(r'[ +\-]?\d+(?:\.\d+[Ee][\+\-]\d\d)?')

    # Go through each line, extract values, then yield them one by one
    for line in fh:
        matches = pattern.findall(line)
        for match in matches:
            if '.' in match:
                yield float(match)
            else:
                yield int(match)




"""
Routines to find critical points (O- and X-points)

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""


from scipy import interpolate
from numpy import zeros, shape
from numpy.linalg import inv
from numpy import dot, linspace, argmax, argmin, abs, clip, sin, cos, pi, amax, arctan2, sqrt, sum
import numpy as np
from warnings import warn


def find_critical(R, Z, psi,discard_xpoints=False):
    """
    Find critical points

    Inputs
    ------
    
    R - R(nr, nz) 2D array of major radii
    Z - Z(nr, nz) 2D array of heights
    psi - psi(nr, nz) 2D array of psi values

    Returns
    -------
    
    Two lists of critical points

    opoint, xpoint

    Each of these is a list of tuples with (R, Z, psi) points

    The first tuple is the primary O-point (magnetic axis)
    and primary X-point (separatrix)
    
    """

    # Get a spline interpolation function
    f = interpolate.RectBivariateSpline(R[:,0], Z[0,:], psi)

    # Find candidate locations, based on minimising Bp^2
    Bp2 = (f(R,Z,dx=1,grid=False)**2 + f(R,Z,dy=1,grid=False)**2) / R**2

    # Get grid resolution, which determines a reasonable tolerance
    # for the Newton iteration search area
    dR = R[1,0] - R[0,0]
    dZ = Z[0,1] - Z[0,0]
    radius_sq = 9*(dR**2 + dZ**2)
    
    # Find local minima

    J = zeros([2,2])

    xpoint = []
    opoint = []
    
    nx,ny = Bp2.shape
    for i in range(2,nx-2):
        for j in range(2,ny-2):
            if ( (Bp2[i,j] < Bp2[i+1,j+1])
                 and (Bp2[i,j] < Bp2[i+1,j])
                 and (Bp2[i,j] < Bp2[i+1,j-1])
                 and (Bp2[i,j] < Bp2[i-1,j+1])
                 and (Bp2[i,j] < Bp2[i-1,j])
                 and (Bp2[i,j] < Bp2[i-1,j-1])
                 and (Bp2[i,j] < Bp2[i,j+1])
                 and (Bp2[i,j] < Bp2[i,j-1]) ):

                # Found local minimum

                R0 = R[i,j]
                Z0 = Z[i,j]
                
                # Use Newton iterations to find where
                # both Br and Bz vanish
                R1 = R0
                Z1 = Z0

                count = 0
                while True:
                
                    Br = -f(R1, Z1, dy=1, grid=False)/R1
                    Bz = f(R1, Z1, dx=1, grid=False)/R1

                    if Br**2 + Bz**2 < 1e-6:
                        # Found a minimum. Classify as either
                        # O-point or X-point

                        # Evaluate D = fxx * fyy - (fxy)^2
                        if False:
                            D = f(R1, Z1, dx=2)[0][0] * f(R1, Z1, dy=2)[0][0] - (f(R1, Z1, dx=1, dy=1)[0][0])**2
                        
                            #print("D0 = %e" % D)
                        
                        if False: #abs(D) < 1:
                            # D small, so need to use another method
                            #print("Small discriminant D (%e)" % (D,))
                            
                            # Try second derivative in index space
                            
                            dR = R[1,0] - R[0,0]
                            dZ = Z[0,1] - Z[0,0]
                            d2dr2 = (psi[i+1,j] - 2.*psi[i,j] + psi[i-1,j])/dR**2
                            d2dz2 = (psi[i,j+1] - 2.*psi[i,j] + psi[i,j-1])/dZ**2
                            d2drdz = ((psi[i+1,j+1] - psi[i+1,j-1])/(2.*dZ) - (psi[i-1,j+1] - psi[i-1,j-1])/(2.*dZ) ) / (2.*dR)
                            D = d2dr2 * d2dz2 - d2drdz**2
                            
                            #print("D1 = %e" % D)
                        
                        if True:
                            dR = R[1,0] - R[0,0]
                            dZ = Z[0,1] - Z[0,0]
                            d2dr2 = (psi[i+2,j] - 2.*psi[i,j] + psi[i-2,j])/(2.*dR)**2
                            d2dz2 = (psi[i,j+2] - 2.*psi[i,j] + psi[i,j-2])/(2.*dZ)**2
                            d2drdz = ((psi[i+2,j+2] - psi[i+2,j-2])/(4.*dZ) - (psi[i-2,j+2] - psi[i-2,j-2])/(4.*dZ) ) / (4.*dR)
                            D = d2dr2 * d2dz2 - d2drdz**2
                            
                            #print("D2 = %e" % D)
                            
                        if D < 0.0:
                            # Found X-point
                            #print("Found X-point at %e, %e (f=%e, D=%e)" % (R1,Z1, f(R1,Z1)[0][0], D) )
                            xpoint.append( (R1,Z1, f(R1,Z1)[0][0]) )
                        else:
                            # Found O-point
                            #print("Found O-point at %e, %e (f=%e, D=%e)" % (R1,Z1, f(R1,Z1)[0][0], D) )
                            opoint.append( (R1,Z1, f(R1,Z1)[0][0]) )
                        break
                        
                    # Jacobian matrix
                    # J = ( dBr/dR, dBr/dZ )
                    #     ( dBz/dR, dBz/dZ )
                
                    J[0,0] = -Br/R1 - f(R1, Z1, dy=1,dx=1)[0][0]/R1
                    J[0,1] = -f(R1, Z1, dy=2)[0][0]/R1
                    J[1,0] = -Bz/R1 + f(R1, Z1, dx=2)/R1
                    J[1,1] = f(R1, Z1, dx=1,dy=1)[0][0]/R1
                
                    d = dot(inv(J), [Br,Bz])

                    R1 = R1 - d[0]
                    Z1 = Z1 - d[1]
                    
                    count += 1
                    # If (R1,Z1) is too far from (R0,Z0) then discard
                    # or if we've taken too many iterations
                    if ((R1 - R0)**2 + (Z1 - Z0)**2 > radius_sq) or (count > 100):
                        # Discard this point
                        break
    
    # Remove duplicates
    def remove_dup(points):
        result = []
        for n, p in enumerate(points):
            dup = False
            for p2 in result:
                if (p[0] - p2[0])**2 + (p[1] - p2[1])**2 < 1e-5:
                    dup = True # Duplicate
                    break
            if not dup:
                result.append(p) # Add to the list
        return result
    
    xpoint = remove_dup(xpoint)
    opoint = remove_dup(opoint)
    
    if len(opoint) == 0:
        # Can't order primary O-point, X-point so return
        print("Warning: No O points found")
        return opoint, xpoint
    
    # Find primary O-point by sorting by distance from middle of domain
    Rmid = 0.5*(R[-1,0] + R[0,0])
    Zmid = 0.5*(Z[0,-1] + Z[0,0])
    opoint.sort(key=lambda x: (x[0] - Rmid)**2 + (x[1] - Zmid)**2)
    
    # Draw a line from the O-point to each X-point. Psi should be
    # monotonic; discard those which are not

    if True: #discard_xpoints:
        Ro,Zo,Po = opoint[0] # The primary O-point
        xpt_keep = []
        for xpt in xpoint:
            Rx, Zx, Px = xpt
            
            rline = linspace(Ro, Rx, num=50)
            zline = linspace(Zo, Zx, num=50)
        
            pline = f(rline, zline, grid=False)

            if Px < Po:
                pline *= -1.0 # Reverse, so pline is maximum at X-point
                
            # Now check that pline is monotonic
            # Tried finding maximum (argmax) and testing
            # how far that is from the X-point. This can go
            # wrong because psi can be quite flat near the X-point
            # Instead here look for the difference in psi
            # rather than the distance in space
            
            maxp = amax(pline)
            if (maxp - pline[-1])/(maxp - pline[0]) > 0.05:
                # More than 5% drop in psi from maximum to X-point
                # -> Discard
                continue
            
            ind = argmin(pline)  # Should be at O-point
            if (rline[ind] - Ro)**2 + (zline[ind] - Zo)**2 > 1e-4:
                # Too far, discard
                continue
            xpt_keep.append(xpt)
        xpoint = xpt_keep

    # Sort X-points by distance to primary O-point in psi space
    psi_axis = opoint[0][2]
    xpoint.sort(key=lambda x: (x[2] - psi_axis)**2)
    
    return opoint, xpoint


def core_mask(R, Z, psi, opoint, xpoint=[], psi_bndry=None):
    """
    Mark the parts of the domain which are in the core
    
    Inputs
    ------

    R[nx,ny] - 2D array of major radius (R) values
    Z[nx,ny] - 2D array of height (Z) values
    psi[nx,ny] - 2D array of poloidal flux

    opoint, xpoint  - Values returned by find_critical

    If psi_bndry is not None, then that is used to find the
    separatrix, not the X-points.

    Returns
    -------
    
    A 2D array [nx,ny] which is 1 inside the core, 0 outside

    """
    
    mask = zeros(psi.shape)
    nx,ny = psi.shape

    # Start and end points
    Ro, Zo, psi_axis = opoint[0]
    if not psi_bndry:
        _, _, psi_bndry = xpoint[0]

    # Normalise psi
    psin = (psi - psi_axis)/(psi_bndry - psi_axis)

    # Need some care near X-points to avoid flood filling through saddle point
    # Here we first set the x-points regions to a value, to block the flood fill
    # then later return to handle these more difficult cases
    # 
    xpt_inds = []
    for rx, zx, _ in xpoint:
        # Find nearest index
        ix = argmin(abs(R[:,0] - rx))
        jx = argmin(abs(Z[0,:] - zx))
        xpt_inds.append((ix,jx))
        # Fill this point and all around with '2'
        for i in np.clip([ix-1,ix,ix+1], 0, nx-1):
            for j in np.clip([jx-1,jx,jx+1], 0, ny-1):
                mask[i,j] = 2

    # Find nearest index to start
    rind = argmin(abs(R[:,0] - Ro))
    zind = argmin(abs(Z[0,:] - Zo))
    
    stack = [(rind, zind)] # List of points to inspect in future
    
    while stack: # Whilst there are any points left
        i, j = stack.pop() # Remove from list
        
        # Check the point to the left (i,j-1)
        if (j > 0) and (psin[i,j-1] < 1.0) and (mask[i,j-1] < 0.5):
            stack.append( (i,j-1) ) 
            
        # Scan along a row to the right
        while True:
            mask[i,j] = 1  # Mark as in the core
            
            if (i < nx-1) and (psin[i+1,j] < 1.0) and (mask[i+1,j] < 0.5):
                stack.append( (i+1,j) )
            if (i > 0) and (psin[i-1,j] < 1.0) and (mask[i-1,j] < 0.5):
                stack.append( (i-1,j) )
                
            if j == ny-1: # End of the row
                break
            if (psin[i,j+1] >= 1.0) or (mask[i,j+1] > 0.5):
                break # Finished this row
            j += 1 # Move to next point along
            
    
    # Now return to X-point locations
    for ix, jx in xpt_inds:
        for i in np.clip([ix-1,ix,ix+1], 0, nx-1):
            for j in np.clip([jx-1,jx,jx+1], 0, ny-1):
                if psin[i,j] < 1.0:
                    mask[i,j] = 1
                else:
                    mask[i,j] = 0
                 
    return mask


def find_psisurface(eq, psifunc, r0,z0, r1,z1, psival=1.0, n=100, axis=None):
    """
    eq      - Equilibrium object
    (r0,z0) - Start location inside separatrix
    (r1,z1) - Location outside separatrix

    n - Number of starting points to use
    """
    # Clip (r1,z1) to be inside domain
    # Shorten the line so that the direction is unchanged
    if abs(r1 - r0) > 1e-6:
        rclip = clip(r1, eq.Rmin, eq.Rmax)
        z1 = z0 + (z1 - z0) * abs( (rclip - r0) / (r1 - r0) )
        r1 = rclip
    
    if abs(z1 - z0) > 1e-6:
        zclip = clip(z1, eq.Zmin, eq.Zmax)
        r1 = r0 + (r1 - r0) * abs( (zclip - z0) / (z1 - z0) )
        z1 = zclip
        
    r = linspace(r0, r1, n)
    z = linspace(z0, z1, n)
        
    if axis is not None:
        axis.plot(r,z)

    pnorm = psifunc(r, z, grid=False)
   
    if hasattr(psival, "__len__"):
        pass
        
    else:
        # Only one value
        ind = argmax(pnorm > psival)

        #Edited by Bhavin 31/07/18
        #Changed 1.0 to psival in f
        #make f gradient to psival surface
        f = (pnorm[ind] - psival)/(pnorm[ind] - pnorm[ind-1])
        
        r = (1. - f) * r[ind] + f * r[ind-1]
        z = (1. - f) * z[ind] + f * z[ind-1]
    
    if axis is not None:
        axis.plot(r,z,'bo')
    
    return r,z


def find_separatrix(eq, opoint=None, xpoint=None, ntheta=20, psi=None, axis=None, psival=1.0):
    """Find the R, Z coordinates of the separatrix for equilbrium
    eq. Returns a tuple of (R, Z, R_X, Z_X), where R_X, Z_X are the
    coordinates of the X-point on the separatrix. Points are equally
    spaced in geometric poloidal angle.

    If opoint, xpoint or psi are not given, they are calculated from eq

    eq - Equilibrium object
    opoint - List of O-point tuples of (R, Z, psi)
    xpoint - List of X-point tuples of (R, Z, psi)
    ntheta - Number of points to find
    psi - Grid of psi on (R, Z)
    axis - A matplotlib axis object to plot points on
    """
    if psi is None:
        psi = eq.psi()

    if (opoint is None) or (xpoint is None):
        opoint, xpoint = find_critical(eq.R, eq.Z, psi)

    psinorm = (psi - opoint[0][2])/(xpoint[0][2] - opoint[0][2])
    
    psifunc = interpolate.RectBivariateSpline(eq.R[:,0], eq.Z[0,:], psinorm)

    r0, z0 = opoint[0][0:2]

    theta_grid = linspace(0, 2*pi, ntheta, endpoint=False)
    dtheta = theta_grid[1] - theta_grid[0]

    # Avoid putting theta grid points exactly on the X-points
    xpoint_theta = arctan2(xpoint[0][0] - r0, xpoint[0][1] - z0)
    # How close in theta to allow theta grid points to the X-point
    TOLERANCE = 1.e-3
    if any(abs(theta_grid - xpoint_theta) < TOLERANCE):
        warn("Theta grid too close to X-point, shifting by half-step")
        theta_grid += dtheta / 2

    isoflux = []
    for theta in theta_grid:
        r, z = find_psisurface(eq, psifunc,
                               r0, z0,
                               r0 + 10.*sin(theta), z0 + 10.*cos(theta),
                               psival=psival,
                               axis=axis)
        isoflux.append((r, z, xpoint[0][0], xpoint[0][1]))

    return isoflux

def find_safety(eq, npsi=1, psinorm=None, ntheta=128, psi=None, opoint=None, xpoint=None,axis=None):
    """Find the safety factor for each value of psi
    Calculates equally spaced flux surfaces. Points on 
    each flux surface are equally paced in poloidal angle
    Performs line integral around flux surface to get q
    
    eq - The equilbrium object
    psinorm flux surface to calculate it for
    npsi - Number of flux surface values to find q for
    ntheta - Number of poloidal points to find it on

    If opoint, xpoint or psi are not given, they are calculated from eq

    returns safety factor for npsi points in normalised psi 
    """


    if psi is None:
        psi = eq.psi()

    if (opoint is None) or (xpoint is None):
        opoint, xpoint = find_critical(eq.R, eq.Z, psi)

    psinormal = (psi - opoint[0][2])/(xpoint[0][2] - opoint[0][2])
    
    psifunc = interpolate.RectBivariateSpline(eq.R[:,0], eq.Z[0,:], psinormal)

    r0, z0 = opoint[0][0:2]
    
    theta_grid = linspace(0, 2*pi, ntheta, endpoint=False)
    dtheta = theta_grid[1] - theta_grid[0]
    
     # Avoid putting theta grid points exactly on the X-points
    xpoint_theta = arctan2(xpoint[0][0] - r0, xpoint[0][1] - z0)
    # How close in theta to allow theta grid points to the X-point
    TOLERANCE = 1.e-3
    
    if any(abs(theta_grid - xpoint_theta) < TOLERANCE):
        warn("Theta grid too close to X-point, shifting by half-step")
        theta_grid += dtheta / 2


    if psinorm is None:
        npsi = 100
        psirange = linspace(0.,1.0,npsi)
    else:
        psirange = psinorm
        npsi = len(psinorm)

    psisurf = zeros([npsi,ntheta,2])


    #Calculate flux surface positions 
    for i in range(npsi):
        psin = psirange[i]
        for j in range(ntheta):
            theta = theta_grid[j]
            r, z = find_psisurface(eq, psifunc,
                                   r0, z0,
                                   r0 + 8.*sin(theta), z0 + 8.*cos(theta),
                                   psival=psin,
                                   axis=axis)
            psisurf[i,j,:] = [r,z]

    
    #Get variables for loop integral around flux surface
    r = psisurf[:,:,0]
    z = psisurf[:,:,1]
    fpol = eq.fpol(psirange[:]).reshape(npsi,1)
    Br = eq.Br(r,z)
    Bz = eq.Bz(r,z)
    Bthe = sqrt(Br**2+Bz**2)


    # Differentiate location w.r.t. index
    dr_di = (np.roll(r, 1, axis=1) - np.roll(r,-1,axis=1))/2.0
    dz_di = (np.roll(z, 1, axis=1) - np.roll(z,-1,axis=1))/2.0

    # Distance between points
    dl = sqrt(dr_di**2 + dz_di**2)

    #Integrand - Btor/(R*Bthe) = Fpol/(R**2*Bthe)
    qint = fpol/(r**2*Bthe)

    #Integral
    q = sum(qint*dl,axis=1)/(2*pi)
 
     
    return  q



"""
Handles reading and writing of Equilibrium objects

Writing is relatively straightforward, but reading requires inferring
the currents in the PF coils

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

from . import _geqdsk
from . import critical
from .equilibrium import Equilibrium
from .machine import Wall
from . import jtor
from . import control
from . import picard
from .gradshafranov import mu0

from scipy import interpolate
from numpy import linspace, amin, amax, reshape, ravel, zeros, argmax, clip, sin, cos, pi
import math
import numpy as np

from scipy.integrate import romb

def write(eq, fh, label=None, oxpoints=None, fileformat=_geqdsk.write):
    """
    Write a GEQDSK equilibrium file, given a FreeGS Equilibrium object
    
    eq - Equilibrium object
    fh - file handle
    
    label - Text label to put in the file
    oxpoints - O- and X-points  (opoint, xpoint) returned by critical.find_critical
          If not given, it will be calculated using critical.find_critical
    """
    # Get poloidal flux
    psi = eq.psi()
    #psi = psi - np.amin(psi)
    # Get size of the grid
    nx,ny = psi.shape

    if oxpoints:
        opoint, xpoint = oxpoints
    else:
        # Find the O- and X-points
        opoint, xpoint = critical.find_critical(eq.R, eq.Z, psi)

    rmin = eq.Rmin
    rmax = eq.Rmax
    zmin = eq.Zmin
    zmax = eq.Zmax

    fvac = eq.fvac() # Vacuum f = R*Bt
    R0 = 1.0 # Reference location
    B0 = fvac / R0 # Reference vacuum toroidal magnetic field
    
    data = {"nx":nx, "ny":ny,
            "rdim":rmax-rmin, # Horizontal dimension in meter of computational box
            "zdim":zmax-zmin, # Vertical dimension in meter of computational box
            "rcentr":R0, # R in meter of vacuum toroidal magnetic field BCENTR
            "bcentr":fvac/R0, # Vacuum magnetic field at rcentr
            "rleft":rmin, # Minimum R in meter of rectangular computational box
            "zmid":0.5*(zmin + zmax)} # Z of center of computational box in meter
            
    
    data["rmagx"], data["zmagx"], data["simagx"] = opoint[0] # magnetic axis

    #Remove Psi magi axis to set it to zero at mag x
    data["sibdry"] = xpoint[0][2]  - data["simagx"] # Psi at boundary
    
    data["cpasma"] = eq.plasmaCurrent() # Plasma current [A]

    psinorm = linspace(0.0, 1.0, nx, endpoint=False) # Does not include separatrix

    data["fpol"] = eq.fpol(psinorm)
    data["pres"] = eq.pressure(psinorm)
    data['ffprime'] = eq.ffprime(psinorm)
    data['pprime'] = eq.pprime(psinorm)

    #Added the pprime and ffprime data into geqdsk.write
    data["fprim"] = eq.ffprime(psinorm)
    data["pprim"] = eq.pprime(psinorm)
    data["psi"] = psi - data["simagx"]
    data["simagx"] = 0.0
    
    qpsi = zeros([nx])
    qpsi[1:] = eq.q(psinorm[1:]) # Exclude axis
    qpsi[0] = qpsi[1]
    data["qpsi"] = qpsi
    
    if eq.tokamak.wall:
        data["rlim"] = eq.tokamak.wall.R
        data["zlim"] = eq.tokamak.wall.Z
    
    # Call fileformat to write the data
    isoflux = np.array(critical.find_separatrix(eq,ntheta=101,opoint=opoint, xpoint=xpoint, psi=psi))

    ind = np.argmin(isoflux[:,1])
    data["rbdry"] = np.roll(isoflux[:,0],-ind)
    data["rbdry"] = np.append(data["rbdry"],data["rbdry"][0])

    data["zbdry"] = np.roll(isoflux[:,1][::-1],-ind)
    data["zbdry"] = np.append(data["zbdry"],data["zbdry"][0])

    
    isoflux = np.array(critical.find_separatrix(eq,ntheta=101,opoint=opoint, xpoint=xpoint, psi=psi, psival=1.1))
    
    data["rlim"] = np.roll(isoflux[:,0][::-1],-ind)
    data["zlim"] = np.roll(isoflux[:,1][::-1],-ind)

    data["rlim"] = np.append(data["rlim"],data["rlim"][0])
    data["zlim"] = np.append(data["zlim"],data["zlim"][0])
    fileformat(data, fh, label=label)

def isPow2(val):
    """
    Returns True if val is a power of 2
    val     - Integer
    """
    return val & (val-1) == 0

def ceilPow2(val):
    """
    Find a power of two greater than or equal to val
    """
    return 2**math.ceil(math.log2(val))

def read(fh, machine, rtol=1e-3, ntheta=8, show=False, axis=None, cocos=1, domain=None, blend=0.0, fit_sol=False):
    """
    Reads a G-EQDSK format file
    
    fh : File handle
    machine : Machine object defining coil locations
    rtol : float
        Relative error in nonlinear solve
    ntheta : integer
        Number of points in poloidal angle on the separatrix
        this is used to constrain the plasma shape
    show : Boolean
        Set to true to show solution in a new figure
    axis : Matplotlib Axis object
       Set to an axis for plotting. Implies show=True
    cocos : integer
        COordinate COnventions. Not fully handled yet,
        only whether psi is divided by 2pi or not.
        if < 10 then psi is divided by 2pi, otherwise not.
    domain : list/tuple of 4 elements
        Sets the (R,Z) domain to solve for
        (Rmin, Rmax, Zmin, Zmax)
    blend : float between 0 and 1
        Weighting of the previous poloidal flux at each step of the
        Picard iteration. The default (0.0) is to use no blending.
        Blending slows convergence, but can stabilise some oscillating
        unstable solutions.
    fit_sol : Boolean
        If False (default) then only the poloidal flux inside the
        separatrix is used to constrain the coil currents.
        This is particularly for reading SCENE input, which is not valid 
        outside the separatrix.
        If True, the whole domain is used in the fitting.
        This is useful if the locations of strike points need to be constrained.

    A nonlinear solve will be performed, using Picard iteration
    
    Returns
    -------

    An Equilibrium object eq. In addition, the following is available:

    eq.control   - The control system
    eq._profiles - The profiles object
    
    """

    if show:
        import matplotlib.pyplot as plt
    
    if fit_sol and domain:
        raise ValueError("Sorry, fit_sol cannot be used with the domain keyword")
    
    if axis is not None:
        show = True

    # Read the data as a Dictionary
    data = _geqdsk.read(fh, cocos=cocos)
    
    # If data contains a limiter, set the machine wall
    if "rlim" in data:
        if len(data["rlim"]) > 3:
            machine.wall = Wall(data["rlim"], data["zlim"])
        else:
            print("Fewer than 3 points given for limiter/wall. Ignoring.")
    

    nx = data["nx"]
    ny = data["ny"]
    psi = data["psi"]
    
    if not (isPow2(nx-1) and isPow2(ny-1)):
        print("Warning: Input grid size %d x %d has sizes which are not 2^n+1" % (nx, ny))
        
        rin = linspace(0, 1, nx)
        zin = linspace(0, 1, ny)
        psi_interp = interpolate.RectBivariateSpline(rin, zin, psi, kx=1, ky=1)
        
        # Ensure that newnx, newny is 2^n + 1
        nx = ceilPow2(nx-1) + 1
        ny = ceilPow2(ny-1) + 1
        print("   => Resizing to %d x %d" % (nx, ny))

        rnew = linspace(0, 1, nx)
        znew = linspace(0, 1, ny)

        # Interpolate onto new grid
        psi = psi_interp(rnew, znew)

        
    # Range of psi normalises psi derivatives
    psi_bndry = data["sibdry"]
    psi_axis = data["simagx"]
    psirange = data["sibdry"] - data["simagx"]
    
    psinorm = linspace(0.0, 1.0, data["nx"], endpoint=True)
    
    # Create a spline fit to pressure, f and f**2
    p_spl = interpolate.InterpolatedUnivariateSpline(psinorm, data["pres"])
    pprime_spl = interpolate.InterpolatedUnivariateSpline(psinorm, data["pres"] / psirange).derivative()
    
    f_spl = interpolate.InterpolatedUnivariateSpline(psinorm, data["fpol"])
    ffprime_spl = interpolate.InterpolatedUnivariateSpline(psinorm, 0.5*data["fpol"]**2/psirange).derivative()

    q_spl = interpolate.InterpolatedUnivariateSpline(psinorm, data["qpsi"])

    # functions to return p, pprime, f and ffprime
    def p_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(p_spl(ravel(psinorm)),psinorm.shape)
        return p_spl(psinorm)

    def f_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(f_spl(ravel(psinorm)),psinorm.shape)
        return f_spl(psinorm)

    def pprime_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(pprime_spl(ravel(psinorm)),psinorm.shape)
        return pprime_spl(psinorm)
    
    def ffprime_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(ffprime_spl(ravel(psinorm)),psinorm.shape)
        return ffprime_spl(psinorm)

    def q_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(q_spl(ravel(psinorm)),psinorm.shape)
        return q_spl(psinorm)
    
    # Create a set of profiles to calculate toroidal current density Jtor
    profiles = jtor.ProfilesPprimeFfprime(pprime_func,
                                          ffprime_func,
                                          data["rcentr"] * data["bcentr"],
                                          p_func=p_func, 
                                          f_func=f_func)
    
    # Calculate normalised psi.
    # 0 = magnetic axis
    # 1 = plasma boundary
    psi_norm = clip((psi - psi_axis)  / (psi_bndry - psi_axis), 0.0, 1.1)

    # Create masking function: 1 inside plasma, 0 outside
    mask = np.ones(psi.shape)
    mask[psi_norm > 1.- 1e-6] = 0.0  # Ignore areas outside the plasma
    
    # Create an Equilibrium object
    eq = Equilibrium(tokamak = machine,
                     Rmin = data["rleft"],
                     Rmax = data["rleft"] + data["rdim"],
                     Zmin = data["zmid"] - 0.5*data["zdim"],
                     Zmax = data["zmid"] + 0.5*data["zdim"],
                     nx=nx, ny=ny         # Number of grid points
                 )
    # Grid spacing
    dR = eq.R[1,0] - eq.R[0,0]
    dZ = eq.Z[0,1] - eq.Z[0,0]
    
    # Toroidal current
    Jtor = eq.R * pprime_func(psi_norm) + ffprime_func(psi_norm)/(eq.R * mu0)
    Jtor *= mask
    
    # Quick calculation of total toroidal current
    print("CURRENT: ", romb(romb(Jtor)) * dR*dZ)
    
    if domain:
        # Change the (R,Z) domain, increasing the resolution if needed
        # to keep the resolution as good or better than the input
        Rmin, Rmax, Zmin, Zmax = domain
        # Calculate approximate grid size needed to keep current resolution
        fnewnx = (Rmax - Rmin)/dR + 1.0
        if fnewnx > nx:
            newnx = int(ceilPow2(fnewnx-1))+1
        fnewny = (Zmax - Zmin)/dZ + 1.0
        if fnewny > ny:
            newny = int(ceilPow2(fnewny-1))+1

        if (newnx != nx) or (newny != ny):
            print("Changing resolution: {} x {}".format(newnx,newny))
            
        # Create an interpolation function for Jtor and the input psi
        Jtor_func = interpolate.RectBivariateSpline(eq.R[:,0], eq.Z[0,:], Jtor)
        psi_func = interpolate.RectBivariateSpline(eq.R[:,0], eq.Z[0,:], psi)
        
        # Create a new Equilibrium object
        # (replacing previous 'eq')
        eq = Equilibrium(tokamak = machine,
                         Rmin = Rmin,
                         Rmax = Rmax,
                         Zmin = Zmin,
                         Zmax = Zmax,
                         nx=nx, ny=ny
                         )
        
        # Interpolate Jtor and psi onto new grid
        Jtor = Jtor_func(eq.R, eq.Z, grid=False)
        psi = psi_func(eq.R, eq.Z, grid=False)

        # Update the mask function by calculating normalised psi
        # on the new grid
        psi_norm = clip((psi - psi_axis)  / (psi_bndry - psi_axis), 0.0, 1.0)

        # Create masking function: 1 inside plasma, 0 outside
        mask = np.ones(psi.shape)
        mask[psi_norm > 1. - 1e-6] = 0.0  # Ignore areas outside the plasma

    # Note: Here we have
    #   eq : Equilibrium object
    #   Jtor : 2D array (nx,ny) Toroidal current density
    #   psi  : 2D array (nx,ny) Input poloidal flux
    #   psi_norm : 2D array (nx,ny) Normalised input poloidal flux
    #   mask : 2D array (nx,ny) 1 inside plasma, 0 outside

    # Perform a linear solve to calculate psi
    # using known Jtor
    eq.solve(profiles, Jtor=Jtor)
    
    print("Plasma current: {0} Amps, input: {1} Amps".format(eq.plasmaCurrent(), data["cpasma"]))
    
    # Identify points to constrain: X-points, O-points and separatrix shape

    if show:
        if axis is None:
            fig = plt.figure()
            axis = fig.add_subplot(111)
        axis.set_aspect('equal')
        axis.set_xlabel("Major radius [m]")
        axis.set_ylabel("Height [m]")

        # Note: psi can be offset by a constant in the input
        axis.contour(eq.R, eq.Z, psi, 50, colors='k')

        # Find all the O- and X-points
        opoint, xpoint = critical.find_critical(eq.R, eq.Z, psi)        
        
        # Put red dots on X-points
        for r,z,_ in xpoint:
            axis.plot(r,z,'ro')

        # Draw separatrix if there is an X-point
        if len(xpoint) > 0:
            sep_contour=axis.contour(eq.R, eq.Z, psi, levels=[xpoint[0][2]], colors='r')
        
    # Find best fit for coil currents
    # First create a control system (see control.py)
    if fit_sol:
        controlsystem = control.ConstrainPsi2D(psi) # Fit entire domain
    else:
        controlsystem = control.ConstrainPsi2D(psi, weights=mask) # Remove SOL from fitting
    # Run control system to find coil currents
    controlsystem(eq)
    
    if show:
        axis.contour(eq.R,eq.Z, eq.psi(), 50, colors='r')
        plt.pause(1)
    
    # Print the coil currents
    machine.printCurrents()

    ####################################################################
    # Refine the equilibrium to ensure consistency 
    # Solve using Picard iteration
    #

    controlsystem = control.ConstrainPsiNorm2D(psi_norm, weights=mask)
    picard.solve(eq,          # The equilibrium to adjust
                 profiles,    # The toroidal current profile function
                 controlsystem, show=show, axis=axis,
                 rtol=rtol, blend=0.5)
    
    print("Plasma current: {0} Amps, input: {1} Amps".format(eq.plasmaCurrent(), data["cpasma"]))
    print("Plasma pressure on axis: {0} Pascals".format(eq.pressure(0.0)))
    machine.printCurrents()
    


    # Attempt to find O- and X-points
    psi = eq.psi()
    opoint, xpoint = critical.find_critical(eq.R, eq.Z, psi)

    if xpoint:
        # Use x-point and o-point constraints because the size of the grid may be changed
        # in which case the 2D psi constraints would fail
        
        # Find the separatrix
        isoflux = critical.find_separatrix(eq, opoint, xpoint, ntheta=ntheta, psi=psi, axis=axis)
        
        # Save the control system to eq
        eq.control = control.constrain(xpoints=xpoint, isoflux=isoflux, gamma=1e-14)

    return eq



"""
Defines class to represent the equilibrium
state, including plasma and coil currents
"""

from numpy import meshgrid, linspace, exp, zeros, nditer, array
from scipy import interpolate
from scipy.integrate import romb, quad # Romberg integration

from .boundary import fixedBoundary, freeBoundary
from .critical import find_separatrix, find_safety

# Operators which define the G-S equation
from .gradshafranov import mu0, GSsparse

# Multigrid solver
from . import multigrid

from . import machine

class Equilibrium:
    """
    Represents the equilibrium state, including
    plasma and coil currents

    Data members
    ------------

    These can be read, but should not be modified directly

    R[nx,ny]
    Z[nx,ny]

    Rmin, Rmax
    Zmin, Zmax

    tokamak - The coils and circuits

    Private data members

    _applyBoundary()
    _solver - Grad-Shafranov elliptic solver

    """

    def __init__(self, tokamak=machine.EmptyTokamak(),
                 Rmin=0.1, Rmax=2.0,
                 Zmin=-1.0, Zmax=1.0,
                 nx=65, ny=65,
                 boundary=freeBoundary,
                 psi=None, current=0.0):
        """Initialises a plasma equilibrium

        Rmin, Rmax  - Range of major radius R [m]
        Zmin, Zmax  - Range of height Z [m]

        nx - Resolution in R. This must be 2^n + 1
        ny - Resolution in Z. This must be 2^m + 1

        boundary - The boundary condition, either freeBoundary or fixedBoundary

        psi - Magnetic flux. If None, use concentric circular flux
              surfaces as starting guess

        current - Plasma current (default = 0.0)
        """

        self.tokamak = tokamak

        self._applyBoundary = boundary

        self.Rmin = Rmin
        self.Rmax = Rmax
        self.Zmin = Zmin
        self.Zmax = Zmax

        self.R_1D = linspace(Rmin, Rmax, nx)
        self.Z_1D = linspace(Zmin, Zmax, ny)
        self.R, self.Z = meshgrid(self.R_1D, self.Z_1D, indexing='ij')

        yymid = 1 - Zmax/(Zmax - Zmin)

        if psi is None:
            # Starting guess for psi
            xx, yy = meshgrid(linspace(0, 1, nx), linspace(0, 1, ny), indexing='ij')
            psi = exp(-((xx - 0.5)**2 + (yy - yymid)**2) / 0.4**2)

            psi[0, :] = 0.0
            psi[:, 0] = 0.0
            psi[-1, :] = 0.0
            psi[:, -1] = 0.0

        self._updatePlasmaPsi(psi)

        # Calculate coil Greens functions. This is an optimisation,
        # used in self.psi() to speed up calculations
        self._pgreen = tokamak.createPsiGreens(self.R, self.Z)

        self._current = current  # Plasma current

        # Create the solver
        generator = GSsparse(Rmin, Rmax, Zmin, Zmax)
        self._solver = multigrid.createVcycle(nx, ny,
                                              generator,
                                              nlevels=1,
                                              ncycle=1,
                                              niter=2,
                                              direct=True)
        
    def setSolverVcycle(nlevels=1, ncycle=1, niter=1, direct=True):
        """
        Creates a new linear solver, based on the multigrid code
    
        nlevels  - Number of resolution levels, including original
        ncycle   - The number of V cycles
        niter    - Number of linear solver (Jacobi) iterations per level
        direct   - Use a direct solver at the coarsest level?
        
        """
        generator = GSsparse(Rmin, Rmax, Zmin, Zmax)
        nx,ny = self.R.shape
        
        self._solver = multigrid.createVcycle(nx, ny,
                                              generator,
                                              nlevels=nlevels,
                                              ncycle=ncycle,
                                              niter=niter,
                                              direct=direct)
        
    def setSolver(solver):
        """
        Sets the linear solver to use. The given object/function must have a __call__ method
        which takes two inputs

        solver(x, b)
        
        where x is the initial guess. This should solve Ax = b, returning the result.

        """
        self._solver = solver

    def callSolver(self, psi, rhs):
        """
        Calls the psi solver, passing the initial guess and RHS arrays
        
        psi   Initial guess for the solution (used if iterative)
        rhs   
        
        Returns
        -------
        
        Solution psi

        """
        return self._solver(psi, rhs)
                        
    def getMachine(self):
        """
        Returns the handle of the machine, including coils
        """
        return self.tokamak
    
    def plasmaCurrent(self):
        """
        Plasma current [Amps]
        """
        return self._current
    
    def plasmaBr(self, R,Z):
        """
        Radial magnetic field due to plasma
        Br = -1/R dpsi/dZ
        """
        return -self.psi_func(R,Z,dy=1, grid=False)/R
        
    def plasmaBz(self, R, Z):
        """
        Vertical magnetic field due to plasma 
        Bz = (1/R) dpsi/dR
        """
        return self.psi_func(R,Z,dx=1, grid=False)/R
        
    def Br(self, R, Z):
        """
        Total radial magnetic field
        """
        return self.plasmaBr(R,Z) + self.tokamak.Br(R,Z)

    def Bz(self, R, Z):
        """
        Total vertical magnetic field
        """
        return self.plasmaBz(R,Z) + self.tokamak.Bz(R,Z)
    
    def psi(self):
        #return self.plasma_psi + self.tokamak.psi(self.R, self.Z)
        return self.plasma_psi + self.tokamak.calcPsiFromGreens(self._pgreen)
        
    def psiRZ(self, R, Z):
        """
        Return poloidal flux psi at given (R,Z) location
        """
        return self.psi_func(R,Z) + self.tokamak.psi(R,Z)

    def fpol(self,psinorm):
        """
        Return f = R*Bt at specified values of normalised psi
        """
        return self._profiles.fpol(psinorm)
    
    def fvac(self):
        """
        Return vacuum f = R*Bt
        """
        return self._profiles.fvac()
    
    def q(self, psinorm):
        """
        Returns safety factor q at specified values of normalised psi
        """
        return find_safety(self,psinorm=psinorm)
        
    def pprime(self, psinorm):
        """
        Return p' at given normalised psi
        """
        return self._profiles.pprime(psinorm)
    
    def ffprime(self, psinorm):
        """
        Return ff' at given normalised psi
        """
        return self._profiles.ffprime(psinorm)
        
    def pressure(self, psinorm, out=None):
        """
        Returns plasma pressure at specified values of normalised psi
        """
        return self._profiles.pressure(psinorm)

    def separatrix(self, ntheta=20):
        """
        Returns an array of ntheta (R, Z) coordinates of the separatrix,
        equally spaced in geometric poloidal angle.
        """
        return array(find_separatrix(self, ntheta=ntheta, psi=self.psi()))[:, 0:2]

    def solve(self, profiles, Jtor=None, psi=None, psi_bndry=None):
        """
        Calculate the plasma equilibrium given new profiles
        replacing the current equilibrium.
        
        This performs the linear Grad-Shafranov solve
        
        profiles  - An object describing the plasma profiles.
                    At minimum this must have methods:
             .Jtor(R, Z, psi)   -> [nx, ny]
             .pprime(psinorm)
             .ffprime(psinorm)
             .pressure(psinorm)
             .fpol(psinorm)

        Jtor : 2D array
            If supplied, specifies the toroidal current at each (R,Z) point
            If not supplied, Jtor is calculated from profiles by finding O,X-points

        psi_bndry  - Poloidal flux to use as the separatrix (plasma boundary)
                     If not given then X-point locations are used.
        """
        
        self._profiles = profiles

        if Jtor is None:
            # Calculate toroidal current density

            if psi is None:
                psi = self.psi()
            Jtor = profiles.Jtor(self.R, self.Z, psi, psi_bndry=psi_bndry)
        
        # Set plasma boundary
        # Note that the Equilibrium is passed to the boundary function
        # since the boundary may need to run the G-S solver (von Hagenow's method)
        self._applyBoundary(self, Jtor, self.plasma_psi)
        
        # Right hand side of G-S equation
        rhs = -mu0 * self.R * Jtor
        
        # Copy boundary conditions
        rhs[0,:] = self.plasma_psi[0,:]
        rhs[:,0] = self.plasma_psi[:,0]
        rhs[-1,:] = self.plasma_psi[-1,:]
        rhs[:,-1] = self.plasma_psi[:,-1]
        
        # Call elliptic solver
        plasma_psi = self._solver(self.plasma_psi, rhs)
        
        self._updatePlasmaPsi(plasma_psi)

        # Update plasma current
        dR = self.R[1,0] - self.R[0,0]
        dZ = self.Z[0,1] - self.Z[0,0]
        self._current = romb(romb(Jtor)) * dR*dZ
        
    def _updatePlasmaPsi(self, plasma_psi):
        """
        Sets the plasma psi data, updates spline interpolation coefficients
        """
        self.plasma_psi = plasma_psi

        # Update spline interpolation
        self.psi_func = interpolate.RectBivariateSpline(self.R[:,0], self.Z[0,:], plasma_psi)
     
    def plot(self, axis=None, show=True, oxpoints=True):
        """
        Plot the equilibrium flux surfaces
        
        axis     - Specify the axis on which to plot
        show     - Call matplotlib.pyplot.show() before returning
        oxpoints - Plot X points as red circles, O points as green circles
    
        """
        from .plotting import plotEquilibrium
        return plotEquilibrium(self, axis=axis, show=show, oxpoints=oxpoints)
    
    def getForces(self):
        """
        Calculate forces on the coils

        Returns a dictionary of coil label -> force
        """
        return self.tokamak.getForces(self)

    def printForces(self):
        """
        Prints a table of forces on coils
        """
        print("Forces on coils")
        def print_forces(forces, prefix=""):
            for label, force in forces.items():
                if isinstance(force, dict):
                    print(prefix + label + " (circuit)")
                    print_forces(force, prefix=prefix + "  ")
                else:
                    print(prefix + label+ " : R = {0:.2f} MN , Z = {1:.2f} MN".format(force[0]*1e-6, force[1]*1e-6))

        print_forces(self.getForces())

def refine(eq):
    """
    Double grid resolution, returning a new equilibrium
    
    """
    # Interpolate the plasma psi
    plasma_psi = multigrid.interpolate(eq.plasma_psi)
    nx, ny = plasma_psi.shape
    
    result = Equilibrium(tokamak=eq.tokamak,
                         Rmin = eq.Rmin,
                         Rmax = eq.Rmax,
                         Zmin = eq.Zmin,
                         Zmax = eq.Zmax,
                         nx=nx, ny=ny)

    result._updatePlasmaPsi(plasma_psi)
    
    if hasattr(eq, "_profiles"):
        result._profiles = eq._profiles

    if hasattr(eq, "control"):
        result.control = eq.control

    return result

def coarsen(eq):
    """
    Reduce grid resolution, returning a new equilibrium
    """
    plasma_psi = multigrid.restrict(eq.plasma_psi)
    nx, ny = plasma_psi.shape
    
    result = Equilibrium(tokamak=eq.tokamak,
                         Rmin = eq.Rmin,
                         Rmax = eq.Rmax,
                         Zmin = eq.Zmin,
                         Zmax = eq.Zmax,
                         nx=nx, ny=ny)

    result._updatePlasmaPsi(plasma_psi)
    
    if hasattr(eq, "_profiles"):
        result._profiles = eq._profiles

    if hasattr(eq, "control"):
        result.control = eq.control

    return result

def newDomain(eq,
              Rmin=None, Rmax=None,
              Zmin=None, Zmax=None,
              nx=None, ny=None):
    """Creates a new Equilibrium, solving in a different domain.
    The domain size (Rmin, Rmax, Zmin, Zmax) and resolution (nx,ny)
    are taken from the input equilibrium eq if not specified.
    """
    if Rmin is None:
        Rmin = eq.Rmin
    if Rmax is None:
        Rmax = eq.Rmax
    if Zmin is None:
        Zmin = eq.Zmin
    if Zmax is None:
        Zmax = eq.Zmax
    if nx is None:
        nx = eq.R.shape[0]
    if ny is None:
        ny = eq.R.shape[0]

    # Create a new equilibrium with the new domain
    result = Equilibrium(tokamak=eq.tokamak,
                         Rmin = Rmin,
                         Rmax = Rmax,
                         Zmin = Zmin,
                         Zmax = Zmax,
                         nx=nx, ny=ny)

    # Calculate the current on the old grid
    profiles = eq._profiles
    Jtor = profiles.Jtor(eq.R, eq.Z, eq.psi())

    # Interpolate Jtor onto new grid
    Jtor_func = interpolate.RectBivariateSpline(eq.R[:,0], eq.Z[0,:], Jtor)
    Jtor_new = Jtor_func(result.R, result.Z, grid=False)

    result._applyBoundary(result, Jtor_new, result.plasma_psi)

    # Right hand side of G-S equation
    rhs = -mu0 * result.R * Jtor_new

    # Copy boundary conditions
    rhs[0,:] = result.plasma_psi[0,:]
    rhs[:,0] = result.plasma_psi[:,0]
    rhs[-1,:] = result.plasma_psi[-1,:]
    rhs[:,-1] = result.plasma_psi[:,-1]

    # Call elliptic solver
    plasma_psi = result._solver(result.plasma_psi, rhs)
        
    result._updatePlasmaPsi(plasma_psi)

    # Solve once more, calculating Jtor using new psi
    result.solve(profiles)
    
    return result


if __name__=="__main__":
    
    # Test the different spline interpolation routines
    
    from numpy import ravel
    import matplotlib.pyplot as plt
    
    import machine
    tokamak = machine.TestTokamak()

    Rmin=0.1
    Rmax=2.0

    eq = Equilibrium(tokamak, Rmin=Rmin, Rmax=Rmax)
    
    import constraints
    xpoints = [(1.2, -0.8),
               (1.2, 0.8)]
    constraints.xpointConstrain(eq, xpoints)
    
    psi = eq.psi()

    tck = interpolate.bisplrep(ravel(eq.R), ravel(eq.Z), ravel(psi))
    spline = interpolate.RectBivariateSpline(eq.R[:,0], eq.Z[0,:], psi)
    f = interpolate.interp2d(eq.R[:,0], eq.Z[0,:],psi, kind='cubic')

    plt.plot(eq.R[:,10], psi[:,10], 'o')

    r = linspace(Rmin, Rmax, 1000)
    z = eq.Z[0,10]
    plt.plot(r, f(r,z), label="f")
    
    plt.plot(r, spline(r,z), label="spline")
    
    plt.plot(r, interpolate.bisplev(r,z, tck), label="bisplev")
    
    plt.legend()
    plt.show()



"""
Functions to impose boundary conditions on psi(plasma)

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

from .gradshafranov import Greens, mu0
from numpy import concatenate, sqrt

from scipy.integrate import romb # Romberg integration

def fixedBoundary(eq, Jtor, psi):
    """
    Set psi=0 on all boundaries
    
    Inputs
    ------
    
    eq    Equilibrium object (not used)
    Jtor  2D array of toroidal current (not used)
    psi   2D array of psi values (modified by call)
    
    Returns
    -------
    None
    """
    psi[0,:] = 0.0
    psi[:,0] = 0.0
    psi[-1,:] = 0.0
    psi[:,-1] = 0.0

def freeBoundary(eq, Jtor, psi):
    """
    Apply a free boundary condition using Green's functions
    
    Note: This method is inefficient because it requires
    an integral over the area of the domain for each point
    on the boundary. 

    Inputs
    ------
    
    eq    Equilibrium object (not used)
    Jtor  2D array of toroidal current (not used)
    psi   2D array of psi values (modified by call)

    Returns
    -------
    
    None
    """
    
    # Get the (R,Z) coordinates from the Equilibrium object
    R = eq.R
    Z = eq.Z

    nx,ny = psi.shape

    dR = R[1,0] - R[0,0]
    dZ = Z[0,1] - Z[0,0]
    
    # List of indices on the boundary
    bndry_indices = concatenate([
        [(x,0) for x in range(nx)],
        [(x,ny-1) for x in range(nx)],
        [(0,y) for y in range(ny)],
        [(nx-1,y) for y in range(ny)]])
    
    for x,y in bndry_indices:
        # Calculate the response of the boundary point
        # to each cell in the plasma domain
        greenfunc = Greens(R, Z, R[x,y], Z[x,y])

        # Prevent infinity/nan by removing (x,y) point
        greenfunc[x,y] = 0.0 
        
        # Integrate over the domain
        psi[x,y] = romb(romb(greenfunc*Jtor))*dR*dZ

def freeBoundaryHagenow(eq, Jtor, psi):
    """
    Apply a free boundary using von Hagenow's method
    
    Inputs
    ------
    
    eq    Equilibrium object (not used)
    Jtor  2D array of toroidal current (not used)
    psi   2D array of psi values (modified by call)

    Returns
    -------
    
    None
    """

    # Get the (R,Z) coordinates from the Equilibrium object
    R = eq.R
    Z = eq.Z

    nx,ny = psi.shape

    dR = R[1,0] - R[0,0]
    dZ = Z[0,1] - Z[0,0]
    
    # solver RHS
    rhs = eq.R * Jtor
    
    # Apply a zero-value boundary condition
    rhs[0,:] = 0.0
    rhs[:,0] = 0.0
    rhs[-1,:] = 0.0
    rhs[:,-1] = 0.0
    
    # Solve with a fixed boundary
    psi_fixed = eq.callSolver(psi, rhs)
    
    # Differentiate at the boundary
    
    #########################################################
    # First order accurate method
    
    #dUdn_L = (psi_fixed[0,:] - psi_fixed[1,:])/dR   # left boundary
    #dUdn_R = (psi_fixed[-1,:] - psi_fixed[-2,:])/dR # Right boundary
    #dUdn_D = (psi_fixed[:,0] - psi_fixed[:,1])/dZ  # Down boundary
    #dUdn_U = (psi_fixed[:,-1] - psi_fixed[:,-2])/dZ  # Upper boundary
    
    # Left down corner
    #dUdn_L[0] = dUdn_D[0] = (psi_fixed[0,0] - psi_fixed[1,1]) / dd
    # Left upper corner
    #dUdn_L[-1] = dUdn_U[0] = (psi_fixed[0,-1] - psi_fixed[1,-2]) / dd
    # Right down corner
    #dUdn_R[0] = dUdn_D[-1] = (psi_fixed[-1,0] - psi_fixed[-2,1]) / dd
    # Right upper corner
    #dUdn_R[-1] = dUdn_U[-1] = (psi_fixed[-1,-1] - psi_fixed[-2,-2]) / dd
    
    #########################################################
    # Second-order one-sided differences at the boundary
    
    dUdn_L = (1.5*psi_fixed[0,:] - 2.*psi_fixed[1,:] + 0.5*psi_fixed[2,:])/dR   # left boundary
    
    dUdn_R = (1.5*psi_fixed[-1,:] - 2.*psi_fixed[-2,:] + 0.5*psi_fixed[-3,:])/dR # Right boundary
    
    dUdn_D = (1.5*psi_fixed[:,0] - 2.*psi_fixed[:,1] + 0.5*psi_fixed[:,2])/dZ  # Down boundary
    
    dUdn_U = (1.5*psi_fixed[:,-1] - 2.*psi_fixed[:,-2] + 0.5*psi_fixed[:,-3])/dZ  # Upper boundary
    
    dd = sqrt(dR**2 + dZ**2) # Diagonal spacing

    # Left down corner 
    dUdn_L[0] = dUdn_D[0] = (1.5*psi_fixed[0,0] - 2.*psi_fixed[1,1] + 0.5*psi_fixed[2,2]) / dd
    
    # Left upper corner
    dUdn_L[-1] = dUdn_U[0] = (1.5*psi_fixed[0,-1] - 2.*psi_fixed[1,-2] + 0.5*psi_fixed[2,-3]) / dd

    # Right down corner
    dUdn_R[0] = dUdn_D[-1] = (1.5*psi_fixed[-1,0] - 2.*psi_fixed[-2,1] + 0.5*psi_fixed[-3,2]) / dd

    # Right upper corner
    dUdn_R[-1] = dUdn_U[-1] = (1.5*psi_fixed[-1,-1] - 2.*psi_fixed[-2,-2] + 0.5*psi_fixed[-3,-3]) / dd
    
    # Now for each point on the boundary perform a loop integral
    
    eps = 1e-2

    # List of indices on the boundary
    # (x index, y index, R change, Z change) 
    # where change in R,Z puts the observation point outside the boundary
    # to avoid the singularity in G(R,R') when R'=R
    
    bndry_indices = concatenate([
        [(x,0,    0.0,-eps) for x in range(nx)],  # Down boundary
        [(x,ny-1, 0.0, eps) for x in range(nx)], # Upper boundary
        [(0,y,   -eps, 0.0) for y in range(ny)],    # Left boundary
        [(nx-1,y, eps, 0.0) for y in range(ny)]]) # Right boundary

    ### APPLY FREE BOUNDARY USING GREENS METHOD
    #freeBoundary(eq, Jtor, psi)

    # Loop through points on boundary
    for x,y,Reps,Zeps in bndry_indices:
        # x and y can be floats here (Python 3.6.4)
        x = int(round(x))
        y = int(round(y))
        
        Rpos = R[x,y]+Reps
        Zpos = Z[x,y]+Zeps

        # Integrate over left boundary

        greenfunc = Greens(R[0,:], Z[0,:], Rpos, Zpos)
        result = romb(greenfunc * dUdn_L / R[0,:]) * dZ
        
        # Integrate over right boundary
        greenfunc = Greens(R[-1,:], Z[-1,:], Rpos, Zpos)
        result += romb(greenfunc * dUdn_R / R[-1,:]) * dZ
        
        # Integrate over down boundary
        greenfunc = Greens(R[:,0], Z[:,0], Rpos, Zpos)
        result += romb(greenfunc * dUdn_D / R[:,0]) * dR
        
        # Integrate over upper boundary
        greenfunc = Greens(R[:,-1], Z[:,-1], Rpos, Zpos)
        result += romb(greenfunc * dUdn_U / R[:,-1]) * dR
        
        ### Compare against Greens method
        #print("%d,%d, %e, %e, %e" % (x,y,psi[x,y], result, result/psi[x,y]))

        psi[x,y] = result
    



"""
Unit tests of G-EQDSK input and output
"""

import unittest
import numpy

try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO

import _geqdsk

class GeqdskTest(unittest.TestCase):

    def test(self):
        
        nx = 65
        ny = 65
        
        # Create a dummy dataset
        data = {"nx":nx, "ny":ny,
                "rdim":2.0,
                "zdim":1.5,
                "rcentr":1.2,
                "bcentr":2.42,
                "rleft": 0.5,
                "zmid":0.1,
                "rmagx": 1.1,
                "zmagx": 0.2,
                "simagx": -2.3,
                "sibdry": 0.21,
                "cpasma": 1234521,
                "fpol":numpy.random.rand(nx),
                "pres":numpy.random.rand(nx),
                "qpsi":numpy.random.rand(nx),
                "psi":numpy.random.rand(nx,ny)
                }

        output = StringIO()

        # Write to string
        _geqdsk.write(data, output)

        # Move to the beginning of the buffer
        output.seek(0)
        
        # Read from string
        data2 = _geqdsk.read(output)

        # Check that data and data2 are the same
        for key in data:
            numpy.testing.assert_allclose(data2[key], data[key])
        

if __name__ == '__main__':
    unittest.main()
    



"""
Plasma control system

Use constraints to adjust coil currents
"""

from numpy import dot, transpose, eye, array
from numpy.linalg import inv
import numpy as np

from scipy import optimize

from . import critical

class constrain(object):
    """
    Adjust coil currents using constraints. To use this class,
    first create an instance by specifying the constraints

    >>> controlsystem = constrain(xpoints = [(1.0, 1.1), (1.0,-1.0)])

    controlsystem will now attempt to create x-points at
    (R,Z) = (1.0, 1.1) and (1.0, -1.1) in any Equilibrium

    >>> controlsystem(eq)

    where eq is an Equilibrium object which is modified by
    the call.
    
    The constraints which can be set are:

    xpoints - A list of X-point (R,Z) locations
    
    isoflux - A list of tuples (R1,Z1, R2,Z2) 
    
    psivals - A list of (R,Z,psi) values

    At least one constraint must be included
    
    gamma - A scalar, minimises the magnitude of the coil currents
    """
    def __init__(self, xpoints=[], gamma=1e-12, isoflux=[], psivals=[]):
        """
        Create an instance, specifying the constraints to apply
        """
        self.xpoints = xpoints
        self.gamma = gamma
        self.isoflux = isoflux
        self.psivals = psivals
        
    def __call__(self, eq):
        """
        Apply constraints to Equilibrium eq
        """
    
        tokamak = eq.getMachine()
    
        constraint_matrix = []
        constraint_rhs = []
        for xpt in self.xpoints:
            # Each x-point introduces two constraints
            # 1) Br = 0

            Br = eq.Br(xpt[0], xpt[1])
            
            # Add currents to cancel out this field
            constraint_rhs.append(-Br)
            constraint_matrix.append(
                tokamak.controlBr(xpt[0], xpt[1]))
            
            # 2) Bz = 0
            
            Bz = eq.Bz(xpt[0], xpt[1])
            
            # Add currents to cancel out this field
            constraint_rhs.append(-Bz)
            constraint_matrix.append(
                tokamak.controlBz(xpt[0], xpt[1]))

        # Constrain points to have the same flux
        for r1,z1, r2,z2 in self.isoflux:
            # Get Psi at (r1,z1) and (r2,z2)
            p1 = eq.psiRZ(r1,z1)
            p2 = eq.psiRZ(r2,z2)
            constraint_rhs.append(p2 - p1)
        
            # Coil responses
            c1 = tokamak.controlPsi(r1,z1)
            c2 = tokamak.controlPsi(r2,z2)
            # Control for the difference between p1 and p2
            c = [ c1val - c2val for c1val, c2val in zip(c1,c2)]
            constraint_matrix.append( c )
    
        # Constrain the value of psi
        for r,z,psi in self.psivals:
            p1 = eq.psiRZ(r,z)
            constraint_rhs.append(psi - p1)

            # Coil responses
            c = tokamak.controlPsi(r,z)
            constraint_matrix.append( c )
            
        if not constraint_rhs:
            raise ValueError("No constraints given")

        # Constraint matrix
        A = array(constraint_matrix)
        b = np.reshape(array(constraint_rhs), (-1,))
    
        # Solve by Tikhonov regularisation
        # minimise || Ax - b ||^2 + ||gamma x ||^2
        #
        # x = (A^T A + gamma^2 I)^{-1}A^T b
        
        # Number of controls (length of x)
        ncontrols = A.shape[1]
        
        # Calculate the change in coil current
        current_change =  dot( inv(dot(transpose(A), A) + self.gamma**2 * eye(ncontrols)), 
                               dot(transpose(A),b))
        print("Current changes: " + str(current_change))
        tokamak.controlAdjust(current_change)

    def plot(self, axis=None, show=True):
        """
        Plots constraints used for coil current control
        
        axis     - Specify the axis on which to plot
        show     - Call matplotlib.pyplot.show() before returning
        
        """
        from .plotting import plotConstraints
        return plotConstraints(self, axis=axis, show=show)
    

def flux_surface(R0, Z0, a, elongation=0.0, triangularity=0.0, indentation=0.0, n=20):
    """
    Generate a set of up-down symmetric isoflux points for a flux
    surface of given shape. Uses the parametrisation given in
      
    J. Manickam, Nucl. Fusion 24 595 (1984)
    
    R0 - Major radius of axis
    Z0 - Height of axis
    a  - Minor radius
    
    elongation - ellipticity-1 = 0 for a circle
    triangularity - 'D' shape
    indentation - Produces a bean shape

    n - Number of points
    
    """
    
    def pos(theta):
        """
        Returns R,Z at given theta
        """
        R = R0 - indentation + (a + indentation*cos(theta))*cos(theta + triangularity*sin(theta))
        Z = Z0 + (1. + elongation) * a * sin(theta)
        return R,Z

    # Poloidal angle values
    thetavals = linspace(0.0, 2*pi, n, endpoint=False)
    
    # Generate the first point, used as a reference for all others
    p0 = pos(0.0)
    
    # All other points are the same flux as p0
    isoflux = [ p0 + pos(theta) for theta in thetavals[1:] ]
    
    return isoflux

class ConstrainPsi2D(object):
    """
    Adjusts coil currents to minimise the square differences
    between psi[R,Z] and a target psi.

    Ignores constant offset differences between psi array
    """
    def __init__(self, target_psi, weights=1.0):
        """
        target_psi : 2D (R,Z) array
            Must be the same size as the equilibrium psi

        weights : float or 2D array of same size as target_psi
            Relative importance of each (R,Z) point in the fitting
            By default every point is equally weighted
            Set points to zero to ignore them in fitting.
           
        """
        # Remove the average so constant offsets are ignored
        self.target_psi = target_psi - np.average(target_psi, weights=weights)
        
        self.weights = weights
        
    def __call__(self, eq):
        """
        Apply constraints to Equilibrium eq
        """
        
        tokamak = eq.getMachine()
        start_currents = tokamak.controlCurrents()
        
        end_currents, _ = optimize.leastsq(self.psi_difference, start_currents, args=(eq,))
        
        tokamak.setControlCurrents(end_currents)
        
    def psi_difference(self, currents, eq):
        """
        Difference between psi from equilibrium with the given currents
        and the target psi
        """
        eq.getMachine().setControlCurrents(currents)
        psi = eq.psi()
        psi_av = np.average(psi, weights=self.weights)
        return ((psi - psi_av - self.target_psi)*self.weights).ravel() # flatten array


class ConstrainPsiNorm2D(object):
    """
    Adjusts coil currents to minimise the square differences
    between normalised psi[R,Z] and a target normalised psi.
    """
    def __init__(self, target_psinorm, weights=1.0):
        """
        target_psinorm : 2D (R,Z) array
            Must be the same size as the equilibrium psi

        weights : float or 2D array of same size as target_psinorm
            Relative importance of each (R,Z) point in the fitting
            By default every point is equally weighted
            Set points to zero to ignore them in fitting.
           
        """
        self.target_psinorm = target_psinorm
        self.weights = weights
        
    def __call__(self, eq):
        """
        Apply constraints to Equilibrium eq
        """
        
        tokamak = eq.getMachine()
        start_currents = tokamak.controlCurrents()
        
        end_currents, _ = optimize.leastsq(self.psinorm_difference, start_currents, args=(eq,))
        
        tokamak.setControlCurrents(end_currents)
        
    def psinorm_difference(self, currents, eq):
        """
        Difference between normalised psi from equilibrium with the given currents
        and the target psinorm
        """
        eq.getMachine().setControlCurrents(currents)
        psi = eq.psi()

        opt, xpt = critical.find_critical(eq.R, eq.Z, psi)
        if not opt:
            raise ValueError("No O-points found!")
        psi_axis = opt[0][2]
        
        if not xpt:
            raise ValueError("No X-points found")
        psi_bndry = xpt[0][2]
        
        # Calculate normalised psi.
        # 0 = magnetic axis
        # 1 = plasma boundary
        psi_norm = (psi - psi_axis)  / (psi_bndry - psi_axis)
        
        return ((psi_norm - self.target_psinorm)*self.weights).ravel() # flatten array



"""
Classes and routines to represent coils and circuits

License
-------

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

from .gradshafranov import Greens, GreensBr, GreensBz, mu0

from numpy import linspace
import numpy as np
import numbers

# We need this for the `label` part of the Circuit dtype for writing
# to HDF5 files. See the following for information:
# http://docs.h5py.org/en/latest/strings.html#how-to-store-text-strings
try:
    import h5py
    has_hdf5 = True
except ImportError:
    has_hdf5 = False


class AreaCurrentLimit:
    """
    Calculate the coil area based on a fixed current density limit
    """
    def __init__(self, current_density = 3.5e9):
        """
        current_density   - Maximum current density in A/m^2
        
        Limits in general depend on the magnetic field
        Typical values Nb3Sn ~ 3.5e9 A/m^2
        https://doi.org/10.1016/0167-899X(86)90010-8
        
        """
        self._current_density = current_density
    
    def __call__(self, coil):
        """
        Return the area in m^2, given a Coil object
        """
        return abs(coil.current * coil.turns) / self._current_density

    
class Coil:
    """
    Represents a poloidal field coil

    public members
    --------------
    
    R, Z - Location of the coil
    current - current in the coil in Amps
    turns   - Number of turns
    control - enable or disable control system
    area    - Cross-section area in m^2

    The total toroidal current carried by the coil is current * turns
    """

    # A dtype for converting to Numpy array and storing in HDF5 files
    dtype = np.dtype([
        (str("R"), np.float64),
        (str("Z"), np.float64),
        (str("current"), np.float64),
        (str("turns"), np.int),
        (str("control"), np.bool),
    ])

    def __init__(self, R, Z, current=0.0, turns=1, control=True, area=AreaCurrentLimit()):
        """
        R, Z - Location of the coil
        
        current - current in each turn of the coil in Amps
        turns   - Number of turns. Total coil current is current * turns
        control - enable or disable control system
        area    - Cross-section area in m^2

        Area can be a fixed value (e.g. 0.025 for 5x5cm coil), or can be specified
        using a function which takes a coil as an input argument.
        To specify a current density limit, use:
      
        area = AreaCurrentLimit(current_density)
        
        where current_density is in A/m^2. The area of the coil will be recalculated
        as the coil current is changed.

        The most important effect of the area is on the coil self-force:
        The smaller the area the larger the hoop force for a given current.
        """
        self.R = R
        self.Z = Z
        
        self.current = current
        self.turns = turns
        self.control = control
        self.area = area

    def psi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z)
        """
        return self.controlPsi(R,Z) * self.current

    def createPsiGreens(self, R, Z):
        """
        Calculate the Greens function at every point, and return
        array. This will be passed back to evaluate Psi in
        calcPsiFromGreens()
        """
        return self.controlPsi(R,Z)

    def calcPsiFromGreens(self, pgreen):
        """
        Calculate plasma psi from Greens functions and current
        """
        return self.current * pgreen

    def Br(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z)
        """
        return self.controlBr(R,Z) * self.current

    def Bz(self, R, Z):
        """
        Calculate vertical magnetic field Bz at (R,Z)
        """
        return self.controlBz(R,Z) * self.current

    def controlPsi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z) due to a unit current
        """
        return Greens(self.R, self.Z, R, Z) * self.turns
        
    def controlBr(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z) due to a unit current
        """
        return GreensBr(self.R,self.Z, R, Z) * self.turns
        
    def controlBz(self, R, Z):
        """
        Calculate vertical magnetic field Bz at (R,Z) due to a unit current
        """
        return GreensBz(self.R,self.Z, R, Z) * self.turns

    def getForces(self, equilibrium):
        """
        Calculate forces on the coils in Newtons
        
        Returns an array of two elements: [ Fr, Fz ]

        
        Force on coil due to its own current:
            Lorentz self‐forces on curved current loops
            Physics of Plasmas 1, 3425 (1998); https://doi.org/10.1063/1.870491
            David A. Garren and James Chen
        """
        current = self.current # current per turn
        total_current = current * self.turns # Total toroidal current

        # Calculate field at this coil due to all other coils
        # and plasma. Need to zero this coil's current
        self.current = 0.0
        Br = equilibrium.Br(self.R, self.Z)
        Bz = equilibrium.Bz(self.R, self.Z)
        self.current = current

        # Assume circular cross-section for hoop (self) force
        minor_radius = np.sqrt(self.area / np.pi)
        
        # Self inductance factor, depending on internal current
        # distribution. 0.5 for uniform current, 0 for surface current
        self_inductance = 0.5

        # Force per unit length.
        # In cgs units f = I^2/(c^2 * R) * (ln(8*R/a) - 1 + xi/2)
        # In SI units f = mu0 * I^2 / (4*pi*R) * (ln(8*R/a) - 1 + xi/2)
        self_fr = (mu0 * total_current**2 / (4.*np.pi*self.R)) * (np.log(8.*self.R/minor_radius) - 1 + self_inductance/2.)
        
        Ltor = 2*np.pi*self.R  # Length of coil
        return np.array([ (total_current * Bz  + self_fr) * Ltor, # Jphi x Bz = Fr, self force always outwards
                          -total_current * Br * Ltor]) # Jphi x Br = - Fz
    
    def __repr__(self):
        return ("Coil(R={0}, Z={1}, current={2}, turns={3}, control={4})"
                .format(self.R, self.Z, self.current, self.turns, self.control))

    def __eq__(self, other):
        return (self.R == other.R
                and self.Z == other.Z
                and self.current == other.current
                and self.turns == other.turns
                and self.control == other.control)

    def __ne__(self, other):
        return not self == other

    def to_numpy_array(self):
        """
        Helper method for writing output
        """
        return np.array((self.R, self.Z, self.current, self.turns, self.control),
                        dtype=self.dtype)

    @classmethod
    def from_numpy_array(cls, value):
        if value.dtype != cls.dtype:
            raise ValueError("Can't create {this} from dtype: {got} (expected: {dtype})"
                             .format(this=type(cls), got=value.dtype, dtype=cls.dtype))
        return Coil(*value[()])

    @property
    def area(self):
        """
        The cross-section area of the coil in m^2
        """
        if isinstance(self._area, numbers.Number):
            assert self._area > 0
            return self._area
        # Calculate using functor
        area = self._area(self)
        assert area > 0
        return area
    
    @area.setter
    def area(self, area):
        self._area = area


class Circuit:
    """
    Represents a collection of coils connected to the same circuit
    
    Public members
    --------------
    
    current  Current in the circuit [Amps]
    control  Use feedback control? [bool]
    """

    # We need a variable-length unicode string in the dtype. The numpy
    # 'S' dtype is not suitable, so we only fallback to it if we don't
    # have HDF5 (in which case, it doesn't matter, but we need
    # something)
    if has_hdf5:
        # Python 2/3 compatibility
        try:
            string_dtype = h5py.special_dtype(vlen=unicode)
        except NameError:
            string_dtype = h5py.special_dtype(vlen=str)
    else:
        string_dtype = np.dtype(str('S'))

    # A dtype for converting to Numpy array and storing in HDF5 files
    dtype = np.dtype([
        (str("label"), string_dtype),
        (str("coil"), Coil.dtype),
        (str("multiplier"), np.float64),
    ])

    def __init__(self, coils, current=0.0, control=True):
        """
        coils - A list [ (label, Coil, multiplier) ]
        """
        
        self.coils = coils
        self.current = current
        self.control = control

    def psi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z)
        """
        psival = 0.0
        for label, coil, multiplier in self.coils:
            coil.current = self.current * multiplier
            psival += coil.psi(R,Z)
        return psival

    def createPsiGreens(self, R, Z):
        """
        Calculate Greens functions
        """
        pgreen = {}
        for label, coil, multiplier in self.coils:
            pgreen[label] = coil.createPsiGreens(R, Z)
        return pgreen

    def calcPsiFromGreens(self, pgreen):
        """
        Calculate psi from Greens functions
        
        """
        psival = 0.0
        for label, coil, multiplier in self.coils:
            coil.current = self.current * multiplier
            psival += coil.calcPsiFromGreens(pgreen[label])
        return psival
        
    def Br(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z)
        """
        result = 0.0
        for label, coil, multiplier in self.coils:
            coil.current = self.current * multiplier
            result += coil.Br(R,Z)
        return result

    def Bz(self, R, Z):
        """
        Calculate vertical magnetic field Bz at (R,Z)
        """
        result = 0.0
        for label, coil, multiplier in self.coils:
            coil.current = self.current * multiplier
            result += coil.Bz(R,Z)
        return result

    def controlPsi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z) due to a unit current
        """
        result = 0.0
        for label, coil, multiplier in self.coils:
            result += multiplier * coil.controlPsi(R, Z)
        return result
        
    def controlBr(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z) due to a unit current
        """
        result = 0.0
        for label, coil, multiplier in self.coils:
            result += multiplier * coil.controlBr(R, Z)
        return result
        
    def controlBz(self, R, Z):
        """
        Calculate vertical magnetic field Bz at (R,Z) due to a unit current
        """
        result = 0.0
        for label, coil, multiplier in self.coils:
            result += multiplier * coil.controlBz(R, Z)
        return result

    def getForces(self, equilibrium):
        """
        Calculate forces on the coils

        Returns a dictionary of coil label -> force
        """
        forces = {}
        for label, coil, multiplier in self.coils:
            forces[label] = coil.getForces(equilibrium)
        return forces
    
    def __repr__(self):
        result = "Circuit(["
        coils = ['("{0}", {1}, {2})'.format(label, coil, multiplier)
                 for label, coil, multiplier in self.coils]
        result += ", ".join(coils)
        return result + "], current={0}, control={1})".format(self.current, self.control)

    def __eq__(self, other):
        return (self.coils == other.coils
                and self.current == other.current
                and self.control == other.control)

    def __ne__(self, other):
        return not self == other

    def to_numpy_array(self):
        """
        Helper method for writing output
        """
        return np.array([(label, coil.to_numpy_array(), multiplier)
                         for label, coil, multiplier in self.coils],
                        dtype=self.dtype)

    @classmethod
    def from_numpy_array(cls, value):
        if value.dtype != cls.dtype:
            raise ValueError("Can't create {this} from dtype: {got} (expected: {dtype})"
                             .format(this=type(cls), got=value.dtype, dtype=cls.dtype))
        # Use the current/control values from the first coil in the circuit
        # Should be consistent!
        return Circuit([(label, Coil(*coil), multiplier) for
                        label, coil, multiplier in value],
                       current=value[0][1]["current"] / value[0]["multiplier"],
                       control=value[0][1]["control"])


class Solenoid:
    """
    Represents a central solenoid

    Public members
    --------------
    
    current - current in each turn
    control - enable or disable control system
    
    """

    # A dtype for converting to Numpy array and storing in HDF5 files
    dtype = np.dtype([
        (str("Rs"), np.float64),
        (str("Zsmin"), np.float64),
        (str("Zsmax"), np.float64),
        (str("Ns"), np.float64),
        (str("current"), np.float64),
        (str("control"), np.bool),
    ])

    def __init__(self, Rs, Zsmin, Zsmax, Ns, current=0.0, control=True):
        """
        Rs - Radius of the solenoid
        Zsmin, Zsmax - Minimum and maximum Z
        Ns - Number of turns
        
        current - current in each turn
        control - enable or disable control system
        """
        self.Rs = Rs
        self.Zsmin = Zsmin
        self.Zsmax = Zsmax
        self.Ns = Ns
        
        self.current = current
        self.control = control
        
    def psi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z)
        """
        return self.controlPsi(R,Z) * self.current

    def createPsiGreens(self, R, Z):
        """
        Calculate Greens functions
        """
        return self.controlPsi(R,Z)

    def calcPsiFromGreens(self, pgreen):
        """
        Calculate psi from Greens functions
        """
        return self.current * pgreen

    def Br(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z)
        """
        return self.controlBr(R,Z) * self.current

    def Bz(self, R, Z):
        """
        Calculate vertical magnetic field Bz at (R,Z)
        """
        return self.controlBz(R,Z) * self.current

    def controlPsi(self, R, Z):
        """
        Calculate poloidal flux at (R,Z) due to a unit current
        
        R and Z should have the same dimensions, but can be multi-dimensional
        Return should have the same shape
        """
        result = 0.0
        for Zs in linspace(self.Zsmin, self.Zsmax, self.Ns):
            result += Greens(self.Rs, Zs, R, Z)
        return result
        
    def controlBr(self, R, Z):
        """
        Calculate radial magnetic field Br at (R,Z) due to a unit current
        """
        result = 0.0
        for Zs in linspace(self.Zsmin, self.Zsmax, self.Ns):
            result += GreensBr(self.Rs, Zs, R, Z)
        return result
        
    def controlBz(self, R, Z):
        """
        Calculate vertical magnetic field Bz at (R,Z) due to a unit current
        """
        result = 0.0
        for Zs in linspace(self.Zsmin, self.Zsmax, self.Ns):
            result += GreensBz(self.Rs, Zs, R, Z)
        return result

    def getForces(self, equilibrium):
        """
        Calculate forces on the solenoid.
        Not currently implemented
        """
        return {}
    
    def __repr__(self):
        return ("Solenoid(Rs={0}, Zsmin={1}, Zsmax={2}, current={3}, Ns={4}, control={5})"
                .format(self.Rs, self.Zsmin, self.Zsmax, self.current, self.Ns, self.control))

    def __eq__(self, other):
        return (self.Rs == other.Rs
                and self.Zsmin == other.Zsmin
                and self.Zsmax == other.Zsmax
                and self.Ns == other.Ns
                and self.current == other.current
                and self.control == other.control)

    def __ne__(self, other):
        return not self == other

    def to_numpy_array(self):
        """
        Helper method for writing output
        """
        return np.array((self.Rs, self.Zsmin, self.Zsmax, self.Ns,
                         self.current, self.control),
                        dtype=self.dtype)

    @classmethod
    def from_numpy_array(cls, value):
        if value.dtype != cls.dtype:
            raise ValueError("Can't create {this} from dtype: {got} (expected: {dtype})"
                             .format(this=type(cls), got=value.dtype, dtype=cls.dtype))
        return Solenoid(*value[()])


class Wall:
    """
    Represents the wall of the device. 
    Consists of an ordered list of (R,Z) points
    """
    def __init__(self, R, Z):
        self.R = R
        self.Z = Z

    def __repr__(self):
        return "Wall(R={R}, Z={Z})".format(R=self.R, Z=self.Z)

    def __eq__(self, other):
        return (self.R == other.R
                and self.Z == other.Z)

    def __ne__(self, other):
        return not self == other
    
    
class Machine:
    """
    Represents the machine (Tokamak), including
    coils and power supply circuits
    
    coils[(label, Coil|Circuit|Solenoid] - List of coils
    
    Note: a list is used rather than a dict, so that the coils
    remain ordered, and so can be updated easily by the control system.
    Instead __getitem__ is implemented to allow access to coils

    """

    def __init__(self, coils, wall=None):
        """
        coils - A list of coils [(label, Coil|Circuit|Solenoid)]
        """
        
        self.coils = coils
        self.wall = wall

    def __repr__(self):
        return ("Machine(coils={coils}, wall={wall})"
                .format(coils=self.coils, wall=self.wall))

    def __eq__(self, other):
        # Other Machine might be equivalent except for order of
        # coils. Assume this doesn't actually matter
        return (sorted(self.coils) == sorted(other.coils)
                and self.wall == other.wall)

    def __ne__(self, other):
        return not self == other

    def __getitem__(self, name):
        for label, coil in self.coils:
            if label == name:
                return coil
        raise KeyError("Machine does not contain coil with label '{0}'".format(name))
        
    def psi(self, R, Z):
        """
        Poloidal flux due to coils
        """
        psi_coils = 0.0
        for label, coil in self.coils:
            psi_coils += coil.psi(R, Z)
        
        return psi_coils

    def createPsiGreens(self, R, Z):
        """
        An optimisation, which pre-computes the Greens functions
        and puts into arrays for each coil. This map can then be
        called at a later time, and quickly return the field
        """
        pgreen = {}
        for label, coil in self.coils:
            pgreen[label] = coil.createPsiGreens(R, Z)
        return pgreen
        
    def calcPsiFromGreens(self, pgreen):
        """
        Uses the object returned by createPsiGreens to quickly
        compute the plasma psi
        """
        psi_coils = 0.0
        for label, coil in self.coils:
            psi_coils += coil.calcPsiFromGreens(pgreen[label])
        return psi_coils

    def Br(self, R, Z):
        """
        Radial magnetic field at given points
        """
        Br = 0.0;    
        for label, coil in self.coils:
            Br += coil.Br(R,Z)
        
        return Br
    
    def Bz(self, R, Z):
        """
        Vertical magnetic field
        """
        Bz = 0.0
        for label, coil in self.coils:
            Bz += coil.Bz(R, Z)
        
        return Bz

    def controlBr(self, R, Z):
        """
        Returns a list of control responses for Br
        at the given (R,Z) location(s).
        """
        return [ coil.controlBr(R,Z)
                 for label, coil in self.coils if coil.control ]

    def controlBz(self, R, Z):
        """
        Returns a list of control responses for Bz
        at the given (R,Z) location(s)
        """
        return [ coil.controlBz(R,Z)
                 for label, coil in self.coils if coil.control ]

    def controlPsi(self, R, Z):
        """
        Returns a list of control responses for psi
        at the given (R,Z) location(s)
        """
        return [ coil.controlPsi(R,Z)
                 for label, coil in self.coils if coil.control]
        

    def controlAdjust(self, current_change):
        """
        Add given currents to the controls.
        Given iterable must be the same length
        as the list returned by controlBr, controlBz
        """
        # Get list of coils being controlled
        controlcoils = [coil for label,coil in self.coils if coil.control]
        
        for coil, dI in zip(controlcoils, current_change):
            # Ensure that dI is a scalar
            coil.current += np.asscalar(dI)

    def controlCurrents(self):
        """
        Return a list of coil currents for the coils being controlled
        """
        return [ coil.current for label, coil in self.coils if coil.control ]

    def setControlCurrents(self, currents):
        """
        Sets the currents in the coils being controlled.
        Input list must be of the same length as the list
        returned by controlCurrents
        """
        controlcoils = [coil for label,coil in self.coils if coil.control]
        for coil, current in zip(controlcoils, currents):
            coil.current = current
            
    def printCurrents(self):
        print("==========================")
        for label, coil in self.coils:
            print(label + " : " + str(coil))
        print("==========================")

    def getForces(self, equilibrium = None):
        """
        Calculate forces on the coils, given the plasma equilibrium.
        If no plasma equilibrium given then the forces due to 
        the coils alone will be calculated.
    
        Returns a dictionary of coil label -> force
        """

        if equilibrium is None:
            equilibrium = self
        
        forces = {}
        for label, coil in self.coils:
            forces[label] = coil.getForces(equilibrium)
        return forces
        
        
def EmptyTokamak():
    """
    Creates a tokamak with no coils
    """
    return Machine([])

def TestTokamak():
    """
    Create a simple tokamak 
    """
    
    coils = [("P1L", Coil(1.0, -1.1)),
             ("P1U", Coil(1.0, 1.1)),
             ("P2L", Coil(1.75, -0.6)),
             ("P2U", Coil(1.75, 0.6))]

    wall = Wall([ 0.75, 0.75,  1.5,  1.8,   1.8,   1.5],   # R
                [-0.85, 0.85, 0.85, 0.25, -0.25, -0.85])   # Z
    
    return Machine(coils, wall)


def DIIID():
    """
    PF coil set from ef20030203.d3d
    Taken from Corsica
    """

    coils = [{"label":"F1A", "R":0.8608, "Z":0.16830, "current":0.0},
             {"label":"F2A", "R":0.8614, "Z":0.50810, "current":0.0},
             {"label":"F3A", "R":0.8628, "Z":0.84910, "current":0.0},
             {"label":"F4A", "R":0.8611, "Z":1.1899 , "current":0.0},
             {"label":"F5A", "R":1.0041, "Z":1.5169 , "current":0.0},
             {"label":"F6A", "R":2.6124, "Z":0.4376 , "current":0.0},
             {"label":"F7A", "R":2.3733, "Z":1.1171 , "current":0.0},
             {"label":"F8A", "R":1.2518, "Z":1.6019 , "current":0.0},
             {"label":"F9A", "R":1.6890, "Z":1.5874 , "current":0.0},
             {"label":"F1B", "R":0.8608, "Z":-0.1737, "current":0.0},
             {"label":"F2B", "R":0.8607, "Z":-0.5135, "current":0.0},
             {"label":"F3B", "R":0.8611, "Z":-0.8543, "current":0.0},
             {"label":"F4B", "R":0.8630, "Z":-1.1957, "current":0.0},
             {"label":"F5B", "R":1.0025, "Z":-1.5169, "current":0.0},
             {"label":"F6B", "R":2.6124, "Z":-0.44376, "current":0.0},
             {"label":"F7B", "R":2.3834, "Z":-1.1171, "current":0.0},
             {"label":"F8B", "R":1.2524, "Z":-1.6027, "current":0.0},
             {"label":"F9B", "R":1.6889, "Z":-1.578, "current":0.0}]
    
    return Machine(coils)

def MAST():
    """
    Mega-Amp Spherical Tokamak. This version has all independent coils 
    so that each is powered by a separate coil
    """
    
    coils = [("P2U", Coil(0.49, 1.76)),
             ("P2L", Coil(0.49, -1.76)),
             ("P3U", Coil(1.1, 1.1)),
             ("P3L", Coil(1.1, -1.1)),
             ("P4U", Coil(1.51, 1.095)),
             ("P4L", Coil(1.51, -1.095)),
             ("P5U", Coil(1.66, 0.52)),
             ("P5L", Coil(1.66, -0.52)),
             ("P6U", Coil(1.5, 0.9)),
             ("P6L", Coil(1.5, -0.9))
             ,("P1", Solenoid(0.15, -1.4, 1.4, 100))
             ] 
    
    return Machine(coils)

def MAST_sym():
    """
    Mega-Amp Spherical Tokamak. This version the upper and lower coils 
    are connected to the same circuits P2 - P6
    """
    coils = [("P2", Circuit( [("P2U", Coil(0.49,  1.76), 1.0),
                              ("P2L", Coil(0.49, -1.76),1.0)] ))
             ,("P3", Circuit( [("P3U", Coil(1.1,  1.1), 1.0),
                              ("P3L", Coil(1.1, -1.1), 1.0)] ))
             ,("P4", Circuit( [("P4U", Coil(1.51,  1.095), 1.0),
                              ("P4L", Coil(1.51, -1.095), 1.0)] ))
             ,("P5", Circuit( [("P5U", Coil(1.66,  0.52), 1.0),
                              ("P5L", Coil(1.66, -0.52), 1.0)] ))
             ,("P6", Circuit( [("P6U", Coil(1.5,  0.9), 1.0),
                               ("P6L", Coil(1.5, -0.9), -1.0)] ))
             ,("P1", Solenoid(0.15, -1.45, 1.45, 100))
         ]
    
    return Machine(coils)

def TCV():
    coils = [
        #             ("A1", Coil(0.422500,0.000000)),
        #             ("B1", Coil(0.445700,-0.936000)),
        #             ("B2", Coil(0.445700,0.936000)),
        #             ("C1", Coil(0.621500,-1.110000)),
        #             ("C2", Coil(0.621500,1.110000)),
        #             ("D1", Coil(1.176500,-1.170000)),
        #             ("D2", Coil(1.176500,1.170000)),
        ("E1", Coil(0.505000,-0.700000)),
        ("E2", Coil(0.505000,-0.500000)),
        ("E3", Coil(0.505000,-0.300000)),
        ("E4", Coil(0.505000,-0.100000)),
        ("E5", Coil(0.505000,0.100000)),
        ("E6", Coil(0.505000,0.300000)),
        ("E7", Coil(0.505000,0.500000)),
        ("E8", Coil(0.505000,0.700000)),
        ("F1", Coil(1.309500,-0.770000)),
        ("F2", Coil(1.309500,-0.610000)),
        ("F3", Coil(1.309500,-0.310000)),
        ("F4", Coil(1.309500,-0.150000)),
        ("F5", Coil(1.309500,0.150000)),
        ("F6", Coil(1.309500,0.310000)),
        ("F7", Coil(1.309500,0.610000)),
        ("F8", Coil(1.309500,0.770000)),
        #             ("G1", Coil(1.098573,-0.651385)),
        #             ("G2", Coil(1.114000,-0.633000)),
        #             ("G3", Coil(1.129427,-0.614615)),
        #             ("G4", Coil(1.129427,0.614615)),
        #             ("G5", Coil(1.114000,0.633000)),
        #             ("G6", Coil(1.098573,0.651385)),
        ("T1", Coil(1.554000,-0.780000)),
        ("T2", Coil(1.717000,-0.780000)),
        ("T3", Coil(1.754000,-0.780000)),
        ("OH", Circuit([("OH1",Solenoid(0.43,-0.93,0.93,100),0.01),
                        ("C1", Coil(0.621500,-1.110000), 1.0),
                        ("C2", Coil(0.621500,1.110000), 1.0),
                        ("D1", Coil(1.176500,-1.170000), 1.0),
                        ("D2", Coil(1.176500,1.170000), 1.0)  ]))
    ]
    
    return Machine(coils)

def MASTU():
    coils = [
        ("Solenoid", Solenoid(0.19475, -1.581, 1.581, 324)),
        ("Pc", Coil(0.067, 0.0, turns=142)),
        ("Px", Circuit([("PxU", Coil(0.2405,  1.2285, turns=44), 1.0),
                        ("PxL", Coil(0.2405, -1.2285, turns=44), 1.0)])),
        ("D1", Circuit([("D1U", Coil(0.381,  1.555, turns=35), 1.0),
                        ("D1L", Coil(0.381, -1.555, turns=35), 1.0)])),
        ("D2", Circuit([("D2U", Coil(0.574,  1.734, turns=23), 1.0),
                        ("D2L", Coil(0.574, -1.734, turns=23), 1.0)])),
        ("D3", Circuit([("D3U", Coil(0.815,  1.980, turns=23), 1.0),
                        ("D3L", Coil(0.815, -1.980, turns=23), 1.0)])),
        ("Dp", Circuit([("DpU", Coil(0.918,  1.501, turns=23), 1.0),
                        ("DpL", Coil(0.918, -1.501, turns=23), 1.0)])),
        ("D5", Circuit([("D5U", Coil(1.900,  1.950, turns=27), 1.0),
                        ("D5L", Coil(1.900, -1.950, turns=27), 1.0)])),
        ("D6", Circuit([("D6U", Coil(1.285,  1.470, turns=23), 1.0),
                        ("D6L", Coil(1.285, -1.470, turns=23), 1.0)])),
        ("D7", Circuit([("D7U", Coil(1.520,  1.470, turns=23), 1.0),
                        ("D7L", Coil(1.520, -1.470, turns=23), 1.0)])),
        ("P4", Circuit([("P4U", Coil(1.500,  1.100, turns=23), 1.0),
                        ("P4L", Coil(1.500, -1.100, turns=23), 1.0)])),
        ("P5", Circuit([("P5U", Coil(1.650,  0.357, turns=23), 1.0),
                        ("P5L", Coil(1.650, -0.357, turns=23), 1.0)])),

        # Vertical control coils wired in opposite directions
        # Two pairs of coils, P6 pair 1 and P6 pair 2
        ("P61", Circuit([("P61U", Coil(1.1975,  1.11175, turns=2),  1.0),
                         ("P61L", Coil(1.1975, -1.11175, turns=2), -1.0)])),
        ("P62", Circuit([("P62U", Coil(1.2575,  1.0575, turns=2),  1.0),
                         ("P62L", Coil(1.2575, -1.0575, turns=2), -1.0)]))
        ]

    rwall = [1.2503, 1.3483, 1.47, 1.47, 1.45, 1.45, 1.3214, 1.1904,
             0.89296, 0.86938, 0.83981, 0.82229, 0.81974, 0.81974,
             0.82734, 0.8548, 0.89017, 0.91974, 0.94066, 1.555, 1.85,
             2, 2, 2, 2, 1.3188, 1.7689, 1.7301, 1.35, 1.09, 1.09,
             0.90576, 0.87889, 0.87889, 0.90717, 0.53948, 0.5112,
             0.50553, 0.53594, 0.5074, 0.4974, 0.5074, 0.4788, 0.4688,
             0.4788, 0.333, 0.333, 0.275, 0.334, 0.261, 0.261, 0.244,
             0.261, 0.261, 0.244, 0.261, 0.261]

    zwall = [1, 0.86, 0.86, 0.81, 0.81, 0.82, 0.82, 1.007, 1.304,
             1.3312, 1.3826, 1.4451, 1.4812, 1.4936, 1.5318, 1.5696,
             1.5891, 1.5936, 1.5936, 1.567, 1.08, 1.08, 1.7, 2.035,
             2.169, 2.169, 1.7189, 1.68, 2.06, 2.06, 2.06, 1.8786,
             1.9055, 1.9055, 1.8772, 1.5095, 1.5378, 1.5321, 1.5017,
             1.4738, 1.4838, 1.4738, 1.4458, 1.4558, 1.4458, 1.303,
             1.1, 1.1, 1.1, 0.502, 0.348, 0.348, 0.348, 0.146, 0.146, 0.146, 0]

    # Concatenate with mirror image in Z,
    # with points in reverse order
    rwall = rwall + rwall[::-1]
    zwall = zwall + [-z for z in zwall[::-1]]
    
    return Machine(coils, Wall(rwall, zwall))


if __name__ == "__main__":
    # Run test case
    
    # Define a machine with a single coil
    coils = [{"R":2.0, "Z":"-1.5", "label":"P1", "current":3.0}]
    tokamak = Machine(coils)
    
    




from . import _divgeo
from . import geqdsk

def write(eq, fh, oxpoints=None):
    """
    Write a DivGeo equilibrium file, given a FreeGS Equilibrium object
    
    eq - Equilibrium object
    fh - file handle
    """

    geqdsk.write(eq, fh, oxpoints=oxpoints, fileformat=_divgeo.write)


    
